name = "bi_tools"

import warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

from bi_tools.BILogger import Logger
from bi_tools.BIParser import Parser
from bi_tools.eztools import timeit
from bi_tools.eztools import backfill
from bi_tools.eztools import col_type_compare_and_match
from bi_tools.eztools import make_directory
from bi_tools.eztools import flex_write
from bi_tools.eztools import flex_read
from bi_tools.eztools import format_for_load
from bi_tools.eztools import redshift_to_snowflake
from bi_tools.eztools import rename_w_snowflake_column_naming_convention
from bi_tools.eztools import remove_garbage
from bi_tools.eztools import create_hash_key
from bi_tools.eztools import wide_to_long
from bi_tools.eztools import long_to_wide
from bi_tools.eztools import timeit



from __future__ import print_function
import os
import os.path
import logging
import sys
import getpass
from datetime import datetime

class log_mode():
    critical = {"CRITICAL":40} #40 write to database, email, and text
    warning = {"WARNING":30} #30 write to database and email
    info = {"INFO":20} #20 write to database
    debug = {"DEBUG":10} #10 print
    notset = {"NOTSET":0} #0 nothing

# snowflake should have a schema called database Logs
# which should be split into different categories of logs
# for example, airflow dag, individual script

#TODO 1: make logger work both on vm and lm. 2: add feature for save to s3
class Logger(object):
    """Logger object for easier logging.

    Args:
        NA
    Returns:
        NA
    Raises:
        NA
    """
    def __init__(self, mode, exc_date=None):
        self.mode = getattr(log_mode, "critical").keys()[0]
        self.mode_score = getattr(log_mode, "critical").values()[0]
        self.username = getpass.getuser()
        self.run_file = os.path.basename(sys.argv[0]).replace(".py", "")
        if exc_date is  None:
            self.exc_date = str(datetime.today().date())
        else:
            self.exc_date = exc_date

        #set up the log files
        self._create_log_file()
        self._config()
        self._set_default_logs()
        self.execution_date()

    def _create_log_file(self):
        #if log directory does not exist, then create so we can write some logs
        log_directory = "/home/{}/logs".format(self.username)
        if not os.path.exists(log_directory):
            os.makedirs(log_directory)

        log_path = "{}/{}_{}.log".format(log_directory,
            self.run_file,
            self.exc_date)
        if os.path.isfile(log_path):
            os.remove(log_path)
        self.logger = logging.getLogger(self.run_file)
        self.handler = logging.FileHandler(log_path)

    def print_log(self, msg):
        """Prints log if the log level exceeds the 'info' level.

        Args:
            msg: log message that you want to print
        Returns:
            NA
        Raises:
            NA
        """
        import sys
        if self.mode_score <= 20:
            print(msg)
        else:
            pass

    def _config(self):
        """Configures the logging settings.

        Args:
            NA
        Returns:
            NA
        Raises:
            NA
        """
        self.logger.setLevel(os.environ.get("LOGLEVEL", self.mode.upper()))
        self.handler.setLevel(os.environ.get("LOGLEVEL", self.mode.upper()))
        format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        formatter = logging.Formatter(format)
        self.handler.setFormatter(formatter)
        self.logger.addHandler(self.handler)

    def custom_log(self, message):
        """Logs a custom message.

        Args:
            message: custom message for logging
        Returns:
            NA
        Raises:
            NA
        """
        self.print_log(message)
        self.logger.info(message)

    def execution_date(self):
        """Logs information about the execution date.

        Args:
            NA
        Returns:
            NA
        Raises:
            NA
        """
        self.print_log(self.execution_date_msg)
        self.logger.info(self.execution_date_msg)

    def ran_info(self):
        self.print_log("{} & {}".format(self.ran_by_msg, self.ran_when_msg))
        self.logger.info("{} & {}".format(self.ran_by_msg, self.ran_when_msg))



import pandas as pd
from datetime import datetime, timedelta

def date_exists_in_db(date, conn, table_name, schema_name, date_var):
    exists = False
    print "checking data for the given date in the db"
    exc_date = str(date)[:10] #in the format of "2018-01-18"
    print exc_date
    query = """select * from {schema}.{table} WHERE
                trunc({date_var}) = '{date}' limit 5;"""\
                    .format(date=exc_date, schema=schema_name,
                            table=table_name, date_var=date_var)
    db_df = conn.sql_dataframe(query)

    if db_df.empty:
        exists = False
    else:
        exists = True

    return exists



from __future__ import print_function
import pandas as pd
import os
import getpass
import argparse
import subprocess
import logging
import boto3
import botocore
from datetime import datetime
from datetime import timedelta
from bi_tools.bi_exceptions import BackfillException
from bi_tools.bi_exceptions import SnowflakeException
from botocore.exceptions import ClientError
from io import BytesIO
import sys
if sys.version_info >= (3,0):
    from io import StringIO
else:
    from StringIO import StringIO
from bi_s3 import S3BI
global username
username = getpass.getuser()

def wide_to_long(df, stubnames, i, j):
    """ Just a wrapper of pandas wide_to_long function
    """
    import pandas as pd
    df= pd.wide_to_long(df, stubnames, i , j)
    return df

def long_to_wide(df, stub, i, j, drop_others=False):
    """ Long to wide panel format. Less flexible but more user friendly than
    pivot tables to get a stata-like reshape.

    This function expects to find a stub column, with values in its rows
    uniquely identified by values in columns i and j. The values in j
    will be applied as suffixes to the stub column to generate a group of
    columns with format Asuffix1, Asuffix2,...AsuffixN.

    Note - will treat all extra columns not included in reshape as additional
    ID columns. The only reason they need to be passed is to ensure that the
    data SHOULD be reshaped, since given enough columns pretty much anything
    can uniquely identify a row.

    Arguments:
    df : DataFrame
         The long-format DataFrame
    stub : str
          The stub name. Contains values in long format. The wide format
          columns will start with this stub name.
    i : str or list
        Columns to use as id variables. Together with j, should uniquely
        identify an observation in a row in stub
    j : str
        Extant column with observations to use as suffix for the stub name.
    drop_others : bool, default=False
                If true, will drop any columns not specified in either i or j.
                Otherwise all columns will be included as additional
                identifier columns.
    """
    if type(i) is str:
        i = [i]
    if isinstance(df.index, pd.core.index.MultiIndex):
        df = df.reset_index()
    # Error Checking
    if df[i + [j]].duplicated().any():
        raise ValueError("i and j don't uniquely identify each row")

    # Perform reshape
    if drop_others:
        df = df[i + [j]]
    else:
        i = [x for x in list(df) if x not in [stub, j]]

	#i = [x for x in i if x not in ["code_system", "extract_type_id"]]

    df = df.set_index(i + [j]).unstack()
    # Ensure all stubs and suffixes are strings and join them to make col names
    cols = pd.Index(df.columns)
    # for each s in each col in cols, e.g. cols = [(s, s1), (s, s2)]
    cols = map(lambda col: map(lambda s: str(s), col), cols)
    cols = [''.join(c) for c in cols]
    # Set columns to the stub+suffix name and remove MultiIndex
    df.columns = cols
    df = df.reset_index()
    return df

def create_hash_key(df, col_list):
    import hashlib
    df["gamma_id"] = ""
    for col in col_list:
        df["gamma_id"] += df[col].astype(str)
    df['gamma_id'] = [hashlib.md5(val).hexdigest() for val in df['gamma_id']]
    return df

def remove_garbage(garbage):
    import gc
    if garbage in globals():
        del globals()[garbage]
    else:
        pass
    gc.collect()

def redshift_to_snowflake(schema, table):
    from bi_db import RedshiftConnection
    from bi_db import SnowflakeConnection
    from datetime import datetime
    rc = RedshiftConnection()
    sc = SnowflakeConnection()
    schema = schema.upper()
    table = table.upper()
    filepath = "s3://pitchbook-snowflake/schema={schema}/table={table}/"\
        "{date}/{schema}_{table}000".format(
            schema=schema,
            table=table,
            date=datetime.strftime(datetime.today(), "%Y-%m-%d")
        )
    rc.unload(database="dev", schema=schema, table=table)
    sc.load(db_name="BUSINESS_INTELLIGENCE",
            schema_name=schema,
            table_name=table,
            filepath=filepath,
            format="csv")
    return "done"

def flex_write(df, save_path, file_type="csv", **kwargs):
    if "s3" in kwargs:
        save_in_s3 = kwargs["s3"]
        if "bucket_name" in kwargs:
            bucket_name = kwargs["bucket_name"]
        else:
            bucket_name = "pitchbook-snowflake"
    else:
        save_in_s3 = False

    if save_in_s3 == True:
        s3bi = S3BI(bucket_name=bucket_name)
        s3bi.write_to_s3(df, save_path)
    else:
        if file_type.lower() == "h5":
            file_type = "hdf"
        flex_write = getattr(df, 'to_{}'.format(file_type))

        if file_type.lower() == "csv":
            flex_write(save_path, index=False)
        elif file_type.lower() == "hdf":
            flex_write(save_path, "w")

        print("{filetype} saved here: {savepath}".format(filetype=file_type,
                                                        savepath=save_path))

def flex_read(load_path, **kwargs):
    if "s3" in kwargs:
        load_from_s3 = kwargs["s3"]
        if "bucket_name" in kwargs:
            bucket_name = kwargs["bucket_name"]
        else:
            bucket_name = "pitchbook-snowflake"
    else:
        load_from_s3 = False

    if "nrows" in kwargs:
        nrows = kwargs["nrows"]
    else:
        nrows = None

    if "usecols" in kwargs:
        usecols = kwargs["usecols"]
    else:
        usecols = None

    if load_from_s3:
        s3bi = S3BI(bucket_name=bucket_name)
        df = s3bi.read_from_s3(load_path, nrows=nrows, usecols=usecols)
    else:
        file_type = load_path.split(".")[-1]
        flex_load = getattr(pd, 'read_{}'.format(file_type))
        df = flex_load(load_path)
        print("Loaded {filetype} from {loadpath}".format(filetype=file_type,
                                                            loadpath=load_path))
    return df

def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print('%r  %2.2f ms' % \
                  (method.__name__, (te - ts) * 1000))
        return result
    return timed

def backfill(filepath, start_date, end_date):
    import bi_db
    shell_path = bi_db.__file__.replace("bi_db", "bi_tools")\
                    .replace(bi_db.__file__.split("/")[-1], "backfill.sh")
    try:
        subprocess.Popen(['chmod', '777', shell_path])
        subprocess.Popen(['sh', shell_path, filepath, start_date, end_date])
    except OSError:
        raise BackfillException("You do not have the shell script in" \
                                "{shellpath}".format(shellpath=shell_path))
    print("backfill job submitted")

def col_type_compare_and_match(df, db_df):
    if len(db_df) > 0:
        df_types = dict(df.dtypes)
        db_types = dict(db_df.dtypes)

        def convert_str_to_unicode(dict_to_convert):
            for key, value in dict_to_convert.items():
                # for now, assume any 'object' variable is unicode
                if value == object:
                    dict_to_convert[key] = type("str")
            return dict_to_convert

        df_types = convert_str_to_unicode(df_types)
        db_types = convert_str_to_unicode(db_types)

        cols_to_convert = []
        sharedKeys = set(df_types.keys()).intersection(db_types.keys())
        for key in sharedKeys:
            if df_types[key] != db_types[key]:
                cols_to_convert.append(key)

        for col in cols_to_convert:
            df[col] = df[col].astype(db_types[col])
    else:
        pass

    return df

def format_for_load(df):
    try:
        datetime_cols = [x for x in df.columns if "_date" in x.lower()]
    except AttributeError as e:
        new_header = df.iloc[0]
        df = df[1:]
        df.columns = new_header
        datetime_cols = [x for x in df.columns if "_date" in x.lower()]
    except:
        raise ValueError("check your s3 object input")
    for col in datetime_cols:
        df[col] = pd.to_datetime(df[col], errors = 'coerce')
    # if all values for the given column is na, then set it to string
    for col in df.columns:
        if df[col].isnull().all():
            df[col] = df[col].astype(str)

    # per sarah's request
    for col in df.columns:
        if "_id" in col.lower():
            df[col] = df[col].fillna(0).astype(int)
            df[col] = df[col].astype(int)
        elif "_min" in col.lower() or "_max" in col.lower():
            df[col] = df[col].astype(float)
        elif "_date" in col.lower() or "timestamp" in col.lower():
            df[col] = pd.to_datetime(df[col])
    return df

def make_directory(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)
    print("{dir} created".format(dir=directory))

def rename_w_snowflake_column_naming_convention(df):
    df.columns = [x.upper().replace(" ","_") for x in df.columns]
    return df

def timeit(func, *args, **kwargs):
    import timeit
    def wrapper(func, *args, **kwargs):
        def wrapped():
            return func(*args, **kwargs)
        return wrapped
    wrapped = wrapper(func, *args, **kwargs)
    return timeit.timeit(wrapped, number=1)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bi_tools",
    version="0.1.7",
    author="BI Data Engineering",
    author_email="jun.kim@pitchbook.com",
    description="PitchBook Business Intelligence Python Toolbox",
    url="https://git.pitchbookdata.com/business-intelligence/bi_tools",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)



import argparse
from datetime import datetime,timedelta

class Parser(object):
    """Parser object for easier system argument parsing.

    Args:
        NA
    Returns:
        NA
    Raises:
        NA
    """
    def __init__(self):
        self.parser = argparse.ArgumentParser(description=
            'Combines test_platform_logs with test_platform_logs_archive')
        self.exc_date, self.mode = self._arg_parser()

    def _arg_parser(self):
        """Parses the arguments.

        Args:
            NA
        Returns:
            NA
        Raises:
            NA
        """

        """
        DATE:
        '2018-06-08'

        LEVELS:
        CRITICAL	50
        ERROR	    40
        WARNING	    30
        INFO	    20
        DEBUG	    10
        NOTSET      0
        """
        today = [str(datetime.today().date())]
        self.parser.add_argument('-d', metavar='N',
                                type=str, nargs='+', default=today,
                                dest='exc_date', help="date of execution")
        self.parser.add_argument('-m', metavar='N',
                        type=str, nargs='+', default=["INFO"],
                        dest='mode', help="logging level")
        args = self.parser.parse_args()
        return args.exc_date[0], args.mode[0]



from BILogger import Logger
from datetime import datetime

class SnowflakeLogger(Logger):
    def __init__(self, schema_name, table_name, **kwargs):
        if "bucket_name" in kwargs:
            self.bucket_name = kwargs["bucket_name"]
        else:
            self.bucket_name = "pitchbook-snowflake"
        self.schema_name = schema_name.upper()
        self.table_name = table_name.upper()
        self.today = datetime.strftime(datetime.today(), "%Y-%m-%d")
        Logger.__init__(self, mode="debug", exc_date=None)

    def move_log_to_s3(self):

    def _create_log_file(self):
        #if log directory does not exist, then create so we can write some logs
        log_directory = "s3://{bucket}/schema={schema}/table={table}/".format(
            bucket=self.bucket_name,
            schema=self.schema_name,
            table=self.table_name
        )
        log_path = "{}/{}_{}.log".format(log_directory,
            self.schema_name,
            self.table_name)
        if os.path.isfile(log_path):
            os.remove(log_path)
        self.logger = logging.getLogger("SNOWFLAKE_LOG")
        self.handler = logging.FileHandler(log_path)
